# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Calibration Details Input Typing """
from dataclasses import dataclass


@dataclass(kw_only=True)
class CalibrationDetailsInput:
    """Calibration Details Input

    Attributes:
        elapsed_time (int | None): elapsed time
        t1 (int | None): last calibrated t1 time
        frequency (int | None): last calibrated frequency
    """

    elapsed_time: int | None = None
    t1: int | None = None  # pylint: disable=invalid-name
    frequency: int | None = None
